==> Patchouli-1.1-25.3-sources.jar: vazkii/patchouli/client/book/template/BookTemplate.java <==
package vazkii.patchouli.client.book.template;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;
import java.util.function.Supplier;

import com.google.gson.annotations.SerializedName;

import net.minecraft.util.ResourceLocation;
import vazkii.patchouli.api.IComponentProcessor;
import vazkii.patchouli.api.IVariableProvider;
import vazkii.patchouli.client.book.BookEntry;
import vazkii.patchouli.client.book.BookPage;
import vazkii.patchouli.client.book.gui.GuiBookEntry;
import vazkii.patchouli.client.book.template.component.ComponentCustom;
import vazkii.patchouli.client.book.template.component.ComponentEntity;
import vazkii.patchouli.client.book.template.component.ComponentFrame;
import vazkii.patchouli.client.book.template.component.ComponentHeader;
import vazkii.patchouli.client.book.template.component.ComponentImage;
import vazkii.patchouli.client.book.template.component.ComponentItemStack;
import vazkii.patchouli.client.book.template.component.ComponentSeparator;
import vazkii.patchouli.client.book.template.component.ComponentText;
import vazkii.patchouli.client.book.template.component.ComponentTooltip;
import vazkii.patchouli.common.book.Book;

public class BookTemplate {
	
	public static final HashMap<String, Class<? extends TemplateComponent>> componentTypes = new HashMap<>();
	
	static {
		registerComponent("text", ComponentText.class);
		registerComponent("item", ComponentItemStack.class);
		registerComponent("image", ComponentImage.class);
		registerComponent("header", ComponentHeader.class);
		registerComponent("separator", ComponentSeparator.class);
		registerComponent("frame", ComponentFrame.class);
		registerComponent("entity", ComponentEntity.class);
		registerComponent("tooltip", ComponentTooltip.class);
		registerComponent("custom", ComponentCustom.class);
	}

	@SerializedName("include")
	List<TemplateInclusion> inclusions = new ArrayList<>();
	List<TemplateComponent> components = new ArrayList<>();
	
	@SerializedName("processor")
	String processorClass;
	
	transient Book book;
	transient TemplateInclusion encapsulation;
	transient IComponentProcessor processor;
	transient boolean compiled = false;
	transient boolean attemptedCreatingProcessor = false;
	
	public static BookTemplate createTemplate(Book book, String type, TemplateInclusion inclusion) {
		ResourceLocation key;
		if(type.contains(":"))
			key = new ResourceLocation(type);
		else key = new ResourceLocation(book.getModNamespace(), type);
		
		Supplier<BookTemplate> supplier = book.contents.templates.get(key);
		if(supplier == null)
			throw new IllegalArgumentException("Template " + key + " does not exist");
		
		BookTemplate template = supplier.get();
		template.book = book;
		template.encapsulation = inclusion;
		
		return template;
	}
	
	public void compile(IVariableProvider variables) {
		if(compiled)
			return;
		
		createProcessor();
		components.removeIf(Objects::isNull);
		
		if(processor != null) {
			IVariableProvider processorVars = variables;
			if(encapsulation != null)
				processorVars = encapsulation.wrapProvider(variables);
			
			try {
				processor.setup(processorVars);
			} catch(Exception e) {
				throw new RuntimeException("Error setting up template processor", e);
			}
		}
		
		for(TemplateInclusion include : inclusions) {
			if(include.template == null || include.template.isEmpty() || include.as == null || include.as.isEmpty())
				throw new IllegalArgumentException("Template inclusion must define both \"template\" and \"as\" fields.");
			
			include.upperMerge(encapsulation);
			include.process(processor);

			BookTemplate template = createTemplate(book, include.template, include);
			template.compile(variables);
			components.addAll(template.components);
		}
		
		for(TemplateComponent c : components)
			c.compile(variables, processor, encapsulation);
		
		compiled = true;
	}
	
	public void build(BookPage page, BookEntry entry, int pageNum) {
		if(compiled)
			components.forEach(c -> {
				c.build(page, entry, pageNum);
			});
	}
	
	public void onDisplayed(BookPage page, GuiBookEntry parent, int left, int top) {
		if(compiled) {
			if(processor != null)
				processor.refresh(parent, left, top);

			components.forEach(c -> c.isVisible = c.getVisibleStatus(processor));
			components.forEach(c -> c.onDisplayed(page, parent, left, top));
		}
	}
	
	public void render(BookPage page, int mouseX, int mouseY, float pticks) { 
		if(compiled)
			components.forEach(c -> {
				if(c.isVisible) 
					c.render(page, mouseX, mouseY, pticks);
			});
	}
	
	public boolean mouseClicked(BookPage page, double mouseX, double mouseY, int mouseButton) {
		if(compiled) {
			for(TemplateComponent c : components)
				if(c.isVisible && c.mouseClicked(page, mouseX, mouseY, mouseButton))
					return true;
		}
		
		return false;
	}
	
	public static void registerComponent(String name, Class<? extends TemplateComponent> clazz) {
		componentTypes.put(name, clazz);
	}
	
	private void createProcessor() {
		if(!attemptedCreatingProcessor) {
			if(processorClass != null && !processorClass.isEmpty()) try {
				Class<?> clazz = Class.forName(processorClass);
				if(clazz != null)
					processor = (IComponentProcessor) clazz.newInstance();
			} catch(Exception e) {
				throw new RuntimeException("Failed to create component processor " + processorClass, e);
			}
			
			attemptedCreatingProcessor = true;
		}
	}

}

==> Patchouli-1.16.4-53.1-sources.jar: vazkii/patchouli/client/book/template/BookTemplate.java <==
package vazkii.patchouli.client.book.template;

import com.google.gson.annotations.SerializedName;
import com.mojang.blaze3d.matrix.MatrixStack;

import net.minecraft.util.ResourceLocation;

import vazkii.patchouli.api.IComponentProcessor;
import vazkii.patchouli.api.IVariableProvider;
import vazkii.patchouli.client.book.BookEntry;
import vazkii.patchouli.client.book.BookPage;
import vazkii.patchouli.client.book.gui.GuiBookEntry;
import vazkii.patchouli.client.book.template.component.*;
import vazkii.patchouli.common.book.Book;

import javax.annotation.Nullable;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Objects;
import java.util.function.Supplier;

public class BookTemplate {

	public static final HashMap<String, Class<? extends TemplateComponent>> componentTypes = new HashMap<>();

	static {
		registerComponent("text", ComponentText.class);
		registerComponent("item", ComponentItemStack.class);
		registerComponent("image", ComponentImage.class);
		registerComponent("header", ComponentHeader.class);
		registerComponent("separator", ComponentSeparator.class);
		registerComponent("frame", ComponentFrame.class);
		registerComponent("entity", ComponentEntity.class);
		registerComponent("tooltip", ComponentTooltip.class);
		registerComponent("custom", ComponentCustom.class);
	}

	@SerializedName("include") List<TemplateInclusion> inclusions = new ArrayList<>();
	List<TemplateComponent> components = new ArrayList<>();

	@SerializedName("processor") String processorClass;

	transient Book book;

	/**
	 * Information about how the parent template is including this one.
	 * Can be null when there is no parent, e.g. the top template at the page level
	 */
	@Nullable transient TemplateInclusion encapsulation;

	transient IComponentProcessor processor;
	transient boolean compiled = false;
	transient boolean attemptedCreatingProcessor = false;

	public static BookTemplate createTemplate(Book book, String type, @Nullable TemplateInclusion inclusion) {
		ResourceLocation key;
		if (type.contains(":")) {
			key = new ResourceLocation(type);
		} else {
			key = new ResourceLocation(book.getModNamespace(), type);
		}

		Supplier<BookTemplate> supplier = book.contents.templates.get(key);
		if (supplier == null) {
			throw new IllegalArgumentException("Template " + key + " does not exist");
		}

		BookTemplate template = supplier.get();
		template.book = book;
		template.encapsulation = inclusion;

		return template;
	}

	public void compile(IVariableProvider variables) {
		if (compiled) {
			return;
		}

		createProcessor();
		components.removeIf(Objects::isNull);

		if (processor != null) {
			IVariableProvider processorVars = variables;
			if (encapsulation != null) {
				processorVars = encapsulation.wrapProvider(variables);
			}

			try {
				processor.setup(processorVars);
			} catch (Exception e) {
				throw new RuntimeException("Error setting up template processor", e);
			}
		}

		for (TemplateInclusion include : inclusions) {
			if (include.template == null || include.template.isEmpty() || include.as == null || include.as.isEmpty()) {
				throw new IllegalArgumentException("Template inclusion must define both \"template\" and \"as\" fields.");
			}

			include.upperMerge(encapsulation);
			include.process(processor);

			BookTemplate template = createTemplate(book, include.template, include);
			template.compile(variables);
			components.addAll(template.components);
		}

		for (TemplateComponent c : components) {
			c.compile(variables, processor, encapsulation);
		}

		compiled = true;
	}

	public void build(BookPage page, BookEntry entry, int pageNum) {
		if (compiled) {
			components.forEach(c -> {
				c.build(page, entry, pageNum);
			});
		}
	}

	public void onDisplayed(BookPage page, GuiBookEntry parent, int left, int top) {
		if (compiled) {
			if (processor != null) {
				processor.refresh(parent, left, top);
			}

			components.forEach(c -> c.isVisible = c.getVisibleStatus(processor));
			components.forEach(c -> c.onDisplayed(page, parent, left, top));
		}
	}

	public void render(MatrixStack ms, BookPage page, int mouseX, int mouseY, float pticks) {
		if (compiled) {
			components.forEach(c -> {
				if (c.isVisible) {
					c.render(ms, page, mouseX, mouseY, pticks);
				}
			});
		}
	}

	public boolean func_231043_a_(BookPage page, double mouseX, double mouseY, int mouseButton) {
		if (compiled) {
			for (TemplateComponent c : components) {
				if (c.isVisible && c.func_231043_a_(page, mouseX, mouseY, mouseButton)) {
					return true;
				}
			}
		}

		return false;
	}

	public static void registerComponent(String name, Class<? extends TemplateComponent> clazz) {
		componentTypes.put(name, clazz);
	}

	private void createProcessor() {
		if (!attemptedCreatingProcessor) {
			if (processorClass != null && !processorClass.isEmpty()) {
				try {
					Class<?> clazz = Class.forName(processorClass);
					processor = (IComponentProcessor) clazz.newInstance();
				} catch (Exception e) {
					throw new RuntimeException("Failed to create component processor " + processorClass, e);
				}
			}

			attemptedCreatingProcessor = true;
		}
	}

}
